/*
 * fiddle with a .sl4 file
 */
#include <stdio.h>
#include <ctype.h>
#include <err.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <string.h>

#define MAXSTIMLIST 20	/* can be increased if needed */
#define STIMLISTGAP1 42	/* bytes between repetitions of name */
#define STIMLISTGAP2 21	/* bytes between 2nd name & data */

void
usage(void) {
	errx(1, "Usage: sl4util [-d|-x|-p][{-l|-L}nnn[:yyy[+sss][@ccc[-[ddd]]]] file.sl4");
}

/*
 * some functions to read in the source .sl4 file
 */
unsigned char *bufsl4;
int szsl4;
int isl4;
#define eof() (isl4>=szsl4)
#define get() (eof()?EOF:bufsl4[isl4++])

void
getsl4(char *sl4) {
	int fd;

	fd = open(sl4, O_RDONLY);
	if (fd < 0) errx(1, "Failed to read %s", sl4);
	szsl4 = lseek(fd, 0, SEEK_END);
	bufsl4 = malloc(szsl4);
	if (!bufsl4) errx(1, "Failed to alloc memory for %s", sl4);
	lseek(fd, 0, SEEK_SET);
	if (szsl4 != read(fd, bufsl4, szsl4)) errx(1, "Failed to read %s", sl4);
	close(fd);
}
/*
 * read a 4-byte number
 */
int
numb(void) {
	int c, x;

	c = get(); if (c == EOF) return (0); x = c;
	c = get(); if (c == EOF) return (0); x |= c<<8;
	c = get(); if (c == EOF) return (0); x |= c<<16;
	c = get(); if (c == EOF) return (0); x |= c<<24;
	return (x);
}
/*
 * convert 4 random bytes to a number
 */
int
cvtn(int i) {
	unsigned char *p = bufsl4 + i;
	return((p[0])|(p[1]<<8)|(p[2]<<16)|(p[3]<<24));
}
/*
 * some print functions for the dump command
 */
void
terpr() {
	static char *nl = "";

	printf("%s%6d", nl, isl4);
	nl = "\n";
}
void
prstr() {
	int len = numb();
	terpr();
	printf(" %02x %02x %02x %02x [%d]",
		len&255, (len>>8)&255, (len>>16)&255, (len>>24)&255, len);
	while (len-- > 0) putchar(get());
}
void
prhex() {
	terpr();
	printf(" %02x", get());
}
/*
 * read a number and see if it is a string
 */
int
isstr(void) {
	int x, c, i, res = 1;
	int back;

	back = isl4;
	x = numb();
	if (x <= 0) {
		res = 0;
	} else {
		for (i = 0; i < x; i++) {
			c = get();
			if (!isprint(c)) {
				res = 0;
				break;
			}
		}
	}
	isl4 = back;
	return(res);
}
/*
 * dump command
 */
void
dump(char *sl4) {
	int x;
	int nhex = 0;

	getsl4(sl4);
	while (!eof()) {
		if (isstr()) {
			prstr();
			nhex = 0;
		} else {
			if (nhex % 16) {
				printf(" %02x", get());
				nhex++;
			} else {
				prhex();
				nhex = 1;
			}
		}
	}
	putchar('\n');
}
/*
 * stimulus list functions
 */
struct stimlist {
	char *name, *file;
	int flag;
	int ucase, skip, bcol, ecol;
} stimlist[MAXSTIMLIST];
int nstimlist;

/* this is in ascending order */
struct patchtab {
	int beg;	/* where to insert it */
	int nxt;	/* the following byte */
	int nstim;
	char **stims;	/* array of nul-terminated stims */
} patchtab[MAXSTIMLIST];
int npatchtab;

#define SBUFINCR BUFSIZ
char *sbuf = NULL;
int sbufsz = 0, nsbuf = 0;
void
addsbuf(int c, int uc) {
	if (nsbuf >= sbufsz) {
		sbufsz += SBUFINCR;
		sbuf = realloc(sbuf, sbufsz);
		if (!sbuf) errx(1, "Out of core");
	}
	if (uc) {
		if (islower(c)) c = toupper(c);
	}
	sbuf[nsbuf++] = c;
}
char *
dupsbuf(void) {
	char *p;

	if (!sbuf) return (NULL);
	p = strdup(sbuf);
	if(!p) errx(1, "Out of core");
	bzero(sbuf, strlen(sbuf));
	nsbuf = 0;
	return(p);
}
void
skipsp(FILE *fp) {
	int c;

	while ((c = getc(fp)) != EOF) {
		if (c != ' ' && c != '\t') {
			ungetc(c, fp);
			return;
		}
	}
}
void
eatnl(FILE *fp) {
	int c;

	while ((c = getc(fp)) != EOF)
		if (c == '\n') break;
}

/*
 * load a stimulus list, processing modifiers
 * and converting _ to space
 */
char **
loadstim(int sl, int ns) {
	FILE *fp;
	struct stimlist *sp = stimlist + sl;
	char **vp;
	int c, i, j;

	if (sl >= nstimlist || sp->name == NULL || sp->file == NULL)
		errx(1, "Tried to load unknown stimlist");
	vp = calloc(ns+1, sizeof(*vp));
	if (!vp) errx(1, "Out of core");
	fp = fopen(sp->file, "r");
	if (!fp) errx(1, "Couldn't open %s", sp->file);
	/*
	 * skip lines if necessary
	 */
	for (i = 0; i < sp->skip; ) {
		c = getc(fp);
		if (c == '\n') i++;
		if (c == EOF) errx(1, "Skipped past EOF in %s", sp->name);
	}
	i = 0;
	while (!feof(fp)) {
		skipsp(fp);
		/* skip to bcol if any */
		for (j = 0; j < sp->bcol - 1; ) {
			c = getc(fp);
			if (c == EOF)
				errx(1, "Not enough lines in %s", sp->name);
			if (c == '\n')
				errx(1, "Not enough columns in %s", sp->name);
			if (c == ' ' || c == '\t') {
				j++;
				skipsp(fp);
				continue;
			}
		}
		/* now copy the required columns */
		while (1) {
			c = getc(fp);
			if (c == EOF) errx(1, "Unexpected EOF");
			if (c == '\n') {
				if (sp->ecol <= 0 || j >= (sp->ecol - 1)) break;
				errx(1, "Not enough columns for specified range in %s", sp->name);
			}
			if (c == ' ' || c == '\t') {
				j++;
				skipsp(fp);
				c = ' ';
				if (sp->ecol > 0 && j > (sp->ecol - 1)) break;
			}
			if (c == '_') c = ' ';
			addsbuf(c, sp->ucase);
		}
		vp[i++] = dupsbuf();
		if (i >= ns) break;
		if (c != '\n') eatnl(fp);
	}
	if (i != ns) errx(1, "Expected %d stims, read %d", ns, i);
	vp[i] = NULL;
	return (vp);
}
/*
 * the stimlist param is of the form name[:path[+nn][@i-j]]
 * the number after the + is how many lines to skip, and the range
 * after the @ is the range of columns to use as the stimulus.
 * a skip <= 0 is no skip, a bcol <= 0 means use all columns, an
 * ecol <= 0 means "to the end of the line".
 * Note that the name argument must be permanent and will be modified.
 */
struct stimlist *
allocstimlist(char *name, int uc) {
	char *p;

	if (nstimlist >= MAXSTIMLIST) errx(1, "Max number of stimlists exceeded");
	p = strchr(name, ':');
	if (p != NULL && *p != 0 && p[1] != 0) {
		stimlist[nstimlist].file = p + 1;
		p = strchr(name, '+');
		if (p) stimlist[nstimlist].skip = atoi(p+1);
		p = strchr(name, '@');
		if (p) {
			stimlist[nstimlist].ecol = stimlist[nstimlist].bcol = atoi(p+1);
			p = strchr(p, '-');
			if (p) stimlist[nstimlist].ecol = atoi(p+1);
		}
		p = strchr(name, '+'); if (p) *p = 0;
		p = strchr(name, '@'); if (p) *p = 0;
		p = strchr(name, ':'); if (p) *p = 0;
	} else {
		if (strchr(name, '+') || strchr(name, '@'))
			errx(1, "'+' and '@' can only appear with ':'");
	}
	if (*name == 0) errx(1, "Null stimlist name");
	stimlist[nstimlist].name = name;
	stimlist[nstimlist].ucase = uc;
	nstimlist++;
}
/*
 * compare an sl4 string with a nul-terminated string
 */
int
eqsl4str(int ix, char *str) {
	int len;

	len = cvtn(ix);
	if (len != strlen(str)) return(0);
	return(!strncmp((char *)bufsl4+ix+4, str, len));
}
/*
 * compare two sl4 strings
 */
int
eqsl4sl4(int ix1, int ix2) {
	int len1, len2;

	len1 = cvtn(ix1);
	len2 = cvtn(ix2);
	if (len1 != len2) return(0);
	return(!strncmp((char *)bufsl4+ix1+4, (char *)bufsl4+ix2+4, len1));
}
/*
 * is the current locate a valid stimlist?
 * returns -1 on failure or the index on success
 */
int
isstimlist(void){
	int i, j, ix;

	if (!isstr()) return(-1);
	j = -1;
	for (i = 0; i < nstimlist; i++) {
		if (!eqsl4str(isl4, stimlist[i].name)) continue;
		ix = isl4+4+cvtn(isl4)+STIMLISTGAP1;
		if (!eqsl4str(ix, stimlist[i].name)) continue;
		j = i; break;
	}
	return(j);
}
/*
 * extract a stimulus list (only one permissible)
 * Also, no path, bcol, ecol, or skip is allowed in the -l parameter
 */
void
extract(char *sl4) {
	int ix;
	int i, j, n;
	int len;

	if (nstimlist != 1
	 || stimlist[0].file != NULL
	 || stimlist[0].skip > 0
	 || stimlist[0].bcol > 0
	 || stimlist[0].ecol > 0
	) errx(1, "There must be exactly one stimlist and no path for extract");
	getsl4(sl4);
	for (; !eof(); get()) {
		if (isstimlist() < 0) continue;
		/* found it */
		isl4 += 2*(4 + cvtn(isl4)) + STIMLISTGAP1 + STIMLISTGAP2;
		n = numb();
		for (i = 0; i < n; i++) {
			if (numb() != 1) errx(1, "Unknown stimflag");
			if (numb() != i) errx(1, "Index mismatch");
			if (!eqsl4sl4(isl4, isl4+4+cvtn(isl4))) errx(1, "Stim mismatch");
			len = numb();
			isl4 += len;
			len = numb();
			for (j = 0; j < len; j++)
				putchar(get());
			putchar('\n');
		}
		return;
	}
	errx(1, "No such stimlist: %s", stimlist[0].name);
}
void
wrnum(int x) {
	putchar(x&255); putchar((x>>8)&255);
	putchar((x>>16)&255); putchar((x>>24)&255);
}
/*
 * patch command--can be one or more stimlists, but each one
 * must have a file associated with it, and may have skip, bcol, and/or ecol.
 */
void
patch(char *sl4) {
	int c, i, j, sl, len;
	char *p;

	if (nstimlist < 1) errx(1, "Must specify at least one stimlist to patch");
	if (isatty(1)) errx(1, "Would write binary data to tty");
	for (i = 0; i < nstimlist; i++)
		if (stimlist[i].file == 0) errx(1, "Must specify a file to import");
	/* first, make a pass and locate each stimlist */
	getsl4(sl4);
	for (; !eof(); get()) {
		sl = isstimlist();
		if (sl < 0) continue;
		stimlist[sl].flag = 1;
		isl4 += 2*(4 + cvtn(isl4)) + STIMLISTGAP1 + STIMLISTGAP2;
		patchtab[npatchtab].nstim = numb();
		patchtab[npatchtab].beg = isl4;
		/* open the file and count the number of stim */
		patchtab[npatchtab].stims = loadstim(sl, patchtab[npatchtab].nstim);
		/* now skip ahead past the stimlist */
		for (i = 0; i < patchtab[npatchtab].nstim; i++) {
			if (numb() != 1) errx(1, "Unknown stimflag");
			if (numb() != i) errx(1, "Index mismatch");
			if (!eqsl4sl4(isl4, isl4+4+cvtn(isl4))) errx(1, "Stim mismatch");
			isl4 += numb();
			isl4 += numb();
		}
		patchtab[npatchtab].nxt = isl4;
		npatchtab++;
	}
	for (i = 0; i < nstimlist; i++)
		if (stimlist[i].flag == 0)
			errx(1, "Couldn't match %s", stimlist[i].name);
	/* ok, now copy out the file with patches */
	isl4 = 0;
	sl = 0;
	while (!eof()) {
		if (sl >= nstimlist) {
			putchar(get());
			continue;
		}
		if (isl4 < patchtab[sl].beg) {
			putchar(get());
			continue;
		}
		for (i = 0; i < patchtab[sl].nstim; i++) {
			p = patchtab[sl].stims[i];
			len = strlen(p);
			wrnum(1);
			wrnum(i);
			wrnum(len);
			for (j = 0; j < len; j++) putchar(p[j]);
			wrnum(len);
			for (j = 0; j < len; j++) putchar(p[j]);
		}
		isl4 = patchtab[sl++].nxt;
	}
}

int
main(int argc, char **argv) {
	int opt, cmd = 'd';

	while ((opt = getopt(argc, argv, "pdxl:L:")) != -1) {
		switch (opt) {
		case 'd': cmd = 'd'; break;
		case 'x': cmd = 'x'; break;
		case 'p': cmd = 'p'; break;
		case 'L': allocstimlist(optarg, 1); break;
		case 'l': allocstimlist(optarg, 0); break;
		case '?':
		default: usage();
		}
	}
	optind--;
	argc -= optind;
	argv += optind;
	if (argc < 2) usage();
	switch (cmd) {
	case 'd': 
		dump(argv[1]);
		break;
	case 'x':
		extract(argv[1]);
		break;
	case 'p':
		patch(argv[1]);
		break;
	}
	exit (0);
}
